/*
 * mod_sleep.c for Apache 2.0
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 *
 *
 * LICENSE
 * -------
 *
 * This source distribution is made freely available and there is no charge
 * for its use, provided you retain this notice, disclaimers, author's
 * copyright, and credits.
 *
 *
 * DISCLAIMER
 * ----------
 *
 * THIS SOFTWARE IS PROVIDE "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO WAY SHALL THE
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define MODULE			"mod_sleep"
#define AUTHOR			"achowe@snert.com"
#define VERSION			"2.1"
#define NDEBUG			1

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include <stdlib.h>

module AP_MODULE_DECLARE_DATA sleep_module;

#define UNSET			-1
#define MERGE(p, c)		(c == UNSET ? p : c)
#define MERGE_PTR(p, c)		((void *) c == (void *) 0 ? p : c)

typedef struct conf {
	unsigned int duration;
} t_conf;

/***********************************************************************
 *** Common Support Routines
 ***********************************************************************/

static const char *
setDuration(const char *arg, unsigned int *number)
{
	long value;
	char *stop;

	value = strtol(arg, (char **) &stop, 10);

	if (value < 0)
		return "Duration must be a positive integer.";

	/* Duration can be expressed in terms of seconds,
	 * minutes, hours, days, and weeks.
	 */
	switch (*stop) {
	default:
		return "Allowed units for duration are: h, m, or s.";
/*
		return "Allowed units for duration are: w, d, h, m, or s.";

	case 'w':
		value *= 7;
	case 'd':
		value *= 24;
*/
	case 'h': case 'H':
		value *= 60;
	case 'm': case 'M':
		value *= 60;
	case 's': case 'S': case '\0':
		break;
	}

	if (INT_MAX < value || value < 0)
		return "Duration too large";

	*number = (unsigned int) value;

	return (const char *) 0;
}

/***********************************************************************
 ***  Create, Init, and Merge Handlers
 ***********************************************************************/

static void *
sleepCreateDir(apr_pool_t *p, char *dir)
{
	t_conf *conf = (t_conf *) apr_palloc(p, sizeof *conf);

	conf->duration = UNSET;

	return (void *) conf;
}

static void *
sleepMergeDir(apr_pool_t *p, void *dad, void *kid)
{
	t_conf *child = (t_conf *) kid;
	t_conf *parent = (t_conf *) dad;

	/* If a value in the child is unset, then inherit the setting of
	 * the parent, which may or may not be the default, otherwise use
	 * the child's value.
	 */
	child->duration = MERGE(parent->duration, child->duration);

	return (void *) child;
}

static int
sleepPostConfig(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	t_conf *conf = ap_get_module_config(s->lookup_defaults, &sleep_module);

        ap_add_version_component(pconf, MODULE "/" VERSION);

	if (conf->duration == UNSET)
		conf->duration = 0;

	return OK;
}

/***********************************************************************
 ***  Phase Handlers
 ***********************************************************************/

static int
sleepFixup(request_rec *r)
{
	t_conf *conf = ap_get_module_config(r->per_dir_config, &sleep_module);

	if (!ap_is_initial_req(r))
		return DECLINED;

	if (conf != (t_conf *) 0) {
#ifndef NDEBUG
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r->server,
			"%s sleeping %u", r->server->server_hostname, conf->duration
		);
#endif
		sleep(conf->duration);
	}

	return DECLINED;
}

static void
sleepHooks(apr_pool_t *p)
{
	ap_hook_post_config(sleepPostConfig, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_fixups(sleepFixup, NULL, NULL, APR_HOOK_MIDDLE);
}

/***********************************************************************
 ***  Configuration Directives
 ***********************************************************************/

/*
 * SleepTime {seconds}
 *
 * Context: server, <virtualhost>, <directory>, <location>
 */
static const char *
SleepTime(cmd_parms *cmd, void *dconf, const char *seconds)
{
	t_conf *conf = dconf;
	const char *msg = setDuration(seconds, &conf->duration);

#ifndef NDEBUG
	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, cmd->server,
		"SleepTime %lu", conf->duration
	);
#endif

	return msg;
}

static command_rec sleepCommands[] = {
	AP_INIT_TAKE1(
		"SleepTime", SleepTime, NULL, RSRC_CONF|ACCESS_CONF,
		"How many seconds to sleep during each request."
	),

	{ NULL }
};

/***********************************************************************
 ***  Module Definition Block
 ***********************************************************************/

module AP_MODULE_DECLARE_DATA sleep_module = {
	STANDARD20_MODULE_STUFF,
	sleepCreateDir,		/* create per-dir    config structures */
	sleepMergeDir,		/* merge  per-dir    config structures */
	NULL,			/* create per-server config structures */
	NULL,			/* merge  per-server config structures */
	sleepCommands,		/* table of config file commands       */
	sleepHooks		/* register hooks */
};
